#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import inspect
import random

import edgedb

from edgedb import compat
from edgedb import _testbase as tb
from edgedb import asyncio_con
from edgedb import asyncio_pool
from edgedb import errors


class TestClient(tb.AsyncQueryTestCase):
    def create_client(self, **kwargs):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.con.dbname
        conargs["timeout"] = 120
        conargs.update(kwargs)

        return edgedb.create_async_pool(**conargs)

    async def test_client_01(self):
        for n in {1, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = await self.create_client(max_size=10)

                async def worker():
                    con = await client.acquire()
                    self.assertEqual(await con.query_single("SELECT 1"), 1)
                    await client.release(con)

                tasks = [worker() for _ in range(n)]
                await asyncio.gather(*tasks)
                await client.aclose()

    async def test_client_02(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                async with self.create_client(max_size=5) as client:

                    async def worker():
                        con = await client.acquire()
                        self.assertEqual(await con.query_single("SELECT 1"), 1)
                        await client.release(con)

                    tasks = [worker() for _ in range(n)]
                    await asyncio.gather(*tasks)

    async def test_client_04(self):
        client = await self.create_client(max_size=1)

        con = await client.acquire()

        # Manual termination of pool connections releases the
        # pool item immediately.
        con.terminate()
        self.assertIsNone(client._impl._holders[0]._con)
        self.assertIsNone(client._impl._holders[0]._in_use)

        con = await client.acquire()
        self.assertEqual(await con.query_single("SELECT 1"), 1)

        await con.aclose()
        self.assertIsNone(client._impl._holders[0]._con)
        self.assertIsNone(client._impl._holders[0]._in_use)
        # Calling release should not hurt.
        await client.release(con)

        client.terminate()

    async def test_client_05(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = await self.create_client(max_size=10)

                async def worker():
                    async with client.acquire() as con:
                        self.assertEqual(await con.query_single("SELECT 1"), 1)

                    self.assertEqual(await client.query('SELECT 1'), [1])
                    self.assertEqual(await client.query_single('SELECT 1'), 1)
                    self.assertEqual(
                        await client.query_json('SELECT 1'), '[1]')
                    self.assertEqual(
                        await client.query_single_json('SELECT 1'), '1')

                tasks = [worker() for _ in range(n)]
                await asyncio.gather(*tasks)
                await client.aclose()

    async def test_client_06(self):
        fut = asyncio.Future()

        async def on_acquire(con):
            fut.set_result(con)

        async with self.create_client(
            max_size=5, on_acquire=on_acquire
        ) as client:
            async with client.acquire() as con:
                pass

        self.assertIs(con, await fut)

    async def test_client_07(self):
        cons = set()

        async def on_acquire(con):
            if con._inner._impl not in cons:  # check underlying connection
                raise RuntimeError("on_connect was not called")

        async def on_connect(con):
            if con._inner._impl in cons:  # check underlying connection
                raise RuntimeError("on_connect was called more than once")
            cons.add(con._inner._impl)

        async def user(pool):
            async with pool.acquire() as con:
                if con._inner._impl not in cons:
                    raise RuntimeError("init was not called")

        async with self.create_client(
            max_size=5, on_connect=on_connect,
            on_acquire=on_acquire,
        ) as client:
            users = asyncio.gather(*[user(client) for _ in range(10)])
            await users

        self.assertEqual(len(cons), 5)

    async def test_client_08(self):
        client = await self.create_client(max_size=1)

        try:
            con = await client.acquire()
            with self.assertRaisesRegex(
                    edgedb.InterfaceError,
                    "does not belong to any connection pool"):
                await client.release(con._inner._impl)
        finally:
            await client.release(con)
            await client.aclose()

    async def test_client_09(self):
        client1 = await self.create_client(max_size=1)

        client2 = await self.create_client(max_size=1)

        try:
            con = await client1.acquire()
            with self.assertRaisesRegex(
                edgedb.InterfaceError, "is not a member"
            ):
                await client2.release(con)
        finally:
            await client1.release(con)

        await client1.aclose()
        await client2.aclose()

    async def test_client_10(self):
        client = await self.create_client(max_size=1)

        con = await client.acquire()
        await client.release(con)
        await client.release(con)

        await client.aclose()

    async def test_client_11(self):
        client = await self.create_client(max_size=1)

        async with client.acquire() as con:
            txn = con.raw_transaction()

        self.assertIn("[released]", repr(con))

        for meth in (
            "query_single",
            "query",
            "execute",
        ):
            with self.assertRaisesRegex(
                edgedb.InterfaceError,
                r"released back to the pool",
            ):
                await getattr(con, meth)("select 1")

        with self.assertRaisesRegex(
            edgedb.InterfaceError,
            r"released back to the pool",
        ):
            await txn.start()

        for meth in ("commit", "rollback"):
            with self.assertRaisesRegex(
                edgedb.InterfaceError,
                r"transaction is not yet started",
            ):
                await getattr(txn, meth)()

        await client.aclose()

    async def test_client_12(self):
        client = await self.create_client(max_size=1)

        async with client.acquire() as con:
            self.assertTrue(isinstance(con, asyncio_con.AsyncIOConnection))
            self.assertFalse(isinstance(con, list))

        await client.aclose()

    async def test_client_13(self):
        client = await self.create_client(max_size=1)

        async with client.acquire() as con:
            self.assertIn("Execute an EdgeQL command", con.execute.__doc__)
            self.assertEqual(con.execute.__name__, "execute")

            self.assertIn(
                str(inspect.signature(con.execute))[1:],
                str(inspect.signature(asyncio_con.AsyncIOConnection.execute)),
            )

        await client.aclose()

    async def test_client_transaction(self):
        client = await self.create_client(max_size=1)

        async for tx in client.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 7*8"), 56)

        await client.aclose()

    async def test_client_retry(self):
        client = await self.create_client(max_size=1)

        async for tx in client.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 7*8"), 56)

        await client.aclose()

    async def test_client_options(self):
        client = await self.create_client(max_size=1)

        client.with_transaction_options(
            edgedb.TransactionOptions(readonly=True))
        client.with_retry_options(
            edgedb.RetryOptions(attempts=1, backoff=edgedb.default_backoff))
        async for tx in client.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 7*8"), 56)

        await client.aclose()

    async def test_client_init_run_until_complete(self):
        client = await self.create_client()
        self.assertIsInstance(client, asyncio_pool.AsyncIOClient)
        await client.aclose()

    async def test_client_exception_in_on_acquire_and_on_connect(self):
        class Error(Exception):
            pass

        async def callback(con):
            nonlocal setup_calls, last_con
            last_con = con
            setup_calls += 1
            if setup_calls > 1:
                cons.append(con)
            else:
                cons.append("error")
                raise Error

        with self.subTest(method="on_acquire"):
            setup_calls = 0
            last_con = None
            cons = []
            async with self.create_client(
                max_size=1, on_acquire=callback
            ) as client:
                with self.assertRaises(Error):
                    await client.acquire()
                self.assertTrue(last_con.is_closed())

                async with client.acquire() as con:
                    self.assertEqual(cons, ["error", con])

        with self.subTest(method="on_connect"):
            setup_calls = 0
            last_con = None
            cons = []
            async with self.create_client(
                max_size=1, on_connect=callback
            ) as client:
                with self.assertRaises(Error):
                    await client.acquire()
                self.assertTrue(last_con.is_closed())

                async with client.acquire() as con:
                    self.assertEqual(await con.query_single("select 1"), 1)
                    self.assertEqual(cons, ["error", con])

    async def test_client_no_acquire_deadlock(self):
        async with self.create_client(
            max_size=1,
        ) as client:

            async with client.acquire() as con:
                has_sleep = await con.query_single("""
                    SELECT EXISTS(
                        SELECT schema::Function FILTER .name = 'sys::_sleep'
                    )
                """)
                if not has_sleep:
                    self.skipTest("No sys::_sleep function")

            async def sleep_and_release():
                async with client.acquire() as con:
                    await con.execute("SELECT sys::_sleep(1)")

            asyncio.ensure_future(sleep_and_release())
            await asyncio.sleep(0.5)

            async with client.acquire() as con:
                await con.query_single("SELECT 1")

    async def test_client_config_persistence(self):
        N = 100
        cons = set()

        class MyConnection(asyncio_pool.PoolConnection):
            async def foo(self):
                return 42

            async def query_single(self, query):
                res = await super().query_single(query)
                return res + 1

        async def test(client):
            async with client.acquire() as con:
                self.assertEqual(await con.query_single("SELECT 1"), 2)
                self.assertEqual(await con.foo(), 42)
                self.assertTrue(isinstance(con, MyConnection))
                cons.add(con)

        async with self.create_client(
            max_size=10,
            connection_class=MyConnection,
        ) as client:

            await asyncio.gather(*[test(client) for _ in range(N)])

        self.assertEqual(len(cons), N)

    async def test_client_connection_methods(self):
        async def test_query(client):
            i = random.randint(0, 20)
            await asyncio.sleep(random.random() / 100)
            r = await client.query("SELECT {}".format(i))
            self.assertEqual(list(r), [i])
            return 1

        async def test_query_single(client):
            i = random.randint(0, 20)
            await asyncio.sleep(random.random() / 100)
            r = await client.query_single("SELECT {}".format(i))
            self.assertEqual(r, i)
            return 1

        async def test_execute(client):
            await asyncio.sleep(random.random() / 100)
            await client.execute("SELECT {1, 2, 3, 4}")
            return 1

        async def run(N, meth):
            async with self.create_client(max_size=10) as client:

                coros = [meth(client) for _ in range(N)]
                res = await asyncio.gather(*coros)
                self.assertEqual(res, [1] * N)

        methods = [
            test_query,
            test_query_single,
            test_execute,
        ]

        with tb.silence_asyncio_long_exec_warning():
            for method in methods:
                with self.subTest(method=method.__name__):
                    await run(200, method)

    async def test_client_handles_transaction_exit_in_asyncgen_1(self):
        client = await self.create_client(max_size=1)

        async def iterate(con):
            async for tx in con.transaction():
                async with tx:
                    for record in await tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            async with client.acquire() as con:
                agen = iterate(con)
                try:
                    async for _ in agen:  # noqa
                        raise MyException()
                finally:
                    await agen.aclose()

        await client.aclose()

    async def test_client_handles_transaction_exit_in_asyncgen_2(self):
        client = await self.create_client(max_size=1)

        async def iterate(con):
            async for tx in con.transaction():
                async with tx:
                    for record in await tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            async with client.acquire() as con:
                iterator = iterate(con)
                try:
                    async for _ in iterator:  # noqa
                        raise MyException()
                finally:
                    await iterator.aclose()

            del iterator

        await client.aclose()

    async def test_client_handles_asyncgen_finalization(self):
        client = await self.create_client(max_size=1)

        async def iterate(tx):
            for record in await tx.query("SELECT {1, 2, 3}"):
                yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            async with client.acquire() as con:
                async for tx in con.transaction():
                    async with tx:
                        agen = iterate(tx)
                        try:
                            async for _ in agen:  # noqa
                                raise MyException()
                        finally:
                            await agen.aclose()

        await client.aclose()

    async def test_client_close_waits_for_release(self):
        client = await self.create_client(max_size=1)

        flag = self.loop.create_future()
        conn_released = False

        async def worker():
            nonlocal conn_released

            async with client.acquire() as connection:
                async with connection.raw_transaction():
                    flag.set_result(True)
                    await asyncio.sleep(0.1)

            conn_released = True

        self.loop.create_task(worker())

        await flag
        await client.aclose()
        self.assertTrue(conn_released)

    async def test_client_close_timeout(self):
        client = await self.create_client(max_size=1)

        flag = self.loop.create_future()

        async def worker():
            async with client.acquire():
                flag.set_result(True)
                await asyncio.sleep(0.5)

        task = self.loop.create_task(worker())

        with self.assertRaises(asyncio.TimeoutError):
            await flag
            await compat.wait_for(client.aclose(), timeout=0.1)

        await task

    async def test_client_expire_connections(self):
        client = await self.create_client(max_size=1)

        con = await client.acquire()
        try:
            await client.expire_connections()
        finally:
            await client.release(con)

        self.assertIsNone(client._impl._holders[0]._con)
        await client.aclose()

    async def test_client_properties(self):
        concurrency = 2

        client = await self.create_client(max_size=concurrency)
        self.assertEqual(client.concurrency, concurrency)
        self.assertEqual(client.concurrency, concurrency)

        async with client.acquire() as _:
            self.assertEqual(client.free_size, concurrency - 1)

        self.assertEqual(client.free_size, concurrency)

        await client.aclose()

    async def _test_connection_broken(self, executor, broken_evt):
        self.assertEqual(await executor.query_single("SELECT 123"), 123)
        broken_evt.set()
        with self.assertRaises(errors.ClientConnectionError):
            await executor.query_single("SELECT 123")
        broken_evt.clear()
        self.assertEqual(await executor.query_single("SELECT 123"), 123)

        tested = False
        async for tx in executor.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 123"), 123)
                if tested:
                    break
                tested = True
                broken_evt.set()
                try:
                    await tx.query_single("SELECT 123")
                except errors.ClientConnectionError:
                    broken_evt.clear()
                    raise
                else:
                    self.fail("ConnectionError not raised!")

    async def test_client_connection_broken(self):
        con_args = self.get_connect_args()
        broken = asyncio.Event()
        done = asyncio.Event()

        async def proxy(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            while True:
                reader = self.loop.create_task(r.read(65536))
                waiter = self.loop.create_task(broken.wait())
                await asyncio.wait(
                    [reader, waiter],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if waiter.done():
                    reader.cancel()
                    w.close()
                    break
                else:
                    waiter.cancel()
                    data = await reader
                    if not data:
                        break
                    w.write(data)

        async def cb(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            ur, uw = await asyncio.open_connection(
                con_args['host'], con_args['port']
            )
            done.clear()
            task = self.loop.create_task(proxy(r, uw))
            try:
                await proxy(ur, w)
            finally:
                try:
                    await task
                finally:
                    done.set()

        server = await asyncio.start_server(
            cb, '127.0.0.1', 0
        )
        port = server.sockets[0].getsockname()[1]
        client = await self.create_client(
            host='127.0.0.1', port=port, max_size=1)
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.con.dbname
        conargs["timeout"] = 120
        conargs["host"] = "127.0.0.1"
        conargs["port"] = port
        conn = await edgedb.async_connect_raw(**conargs)
        try:
            await self._test_connection_broken(conn, broken)
            await self._test_connection_broken(client, broken)
        finally:
            server.close()
            await server.wait_closed()
            await asyncio.wait_for(client.aclose(), 5)
            await asyncio.wait_for(conn.aclose(), 1)
            broken.set()
            await done.wait()

    async def test_client_suggested_concurrency(self):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.con.dbname
        conargs["timeout"] = 120

        client = edgedb.create_async_client(**conargs)

        self.assertEqual(client.concurrency, 1)

        await client.ensure_connected()
        self.assertGreater(client.concurrency, 1)

        await client.aclose()

        client = edgedb.create_async_client(**conargs, concurrency=5)

        self.assertEqual(client.concurrency, 5)

        await client.ensure_connected()
        self.assertEqual(client.concurrency, 5)

        await client.aclose()

    async def test_client_deprecated_pool(self):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.con.dbname
        conargs["timeout"] = 120

        pool = await edgedb.create_async_pool(**conargs)

        await pool.aclose()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


cdef class BaseCodec:

    cdef:
        bytes   tid
        str     name

    cdef inline bytes get_tid(self):
        return self.tid

    cdef encode(self, WriteBuffer buf, object obj)
    cdef decode(self, FRBuffer *buf)

    cdef dump(self, int level=?)


cdef class CodecPythonOverride(BaseCodec):

    cdef:
        BaseCodec codec
        object encoder
        object decoder

    @staticmethod
    cdef BaseCodec new(bytes tid,
                       BaseCodec basecodec,
                       object encoder,
                       object decoder)


cdef class BaseRecordCodec(BaseCodec):

    cdef:
        tuple fields_codecs
        uint64_t encoder_flags

    cdef _check_encoder(self)


cdef class EmptyTupleCodec(BaseCodec):

    cdef:
        object empty_tup


cdef class NullCodec(BaseCodec):
    pass


cdef class BaseNamedRecordCodec(BaseRecordCodec):

    cdef:
        object descriptor


@cython.final
cdef class EdegDBCodecContext(pgproto.CodecContext):

    cdef:
        object _codec



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# flake8: noqa

from ._version import __version__

from edgedb.datatypes.datatypes import (
    Tuple, NamedTuple, EnumValue, RelativeDuration, ConfigMemory
)
from edgedb.datatypes.datatypes import Set, Object, Array, Link, LinkSet

from .abstract import Executor, AsyncIOExecutor

from .asyncio_con import async_connect_raw, AsyncIOConnection
from .asyncio_pool import (
    create_async_client,
    create_async_pool,
    async_connect,
    AsyncIOClient
)

from .blocking_con import connect, BlockingIOConnection
from .options import RetryCondition, IsolationLevel, default_backoff
from .options import RetryOptions, TransactionOptions
from .transaction import Transaction, AsyncIOTransaction

from .errors._base import EdgeDBError, EdgeDBMessage


# The below is generated by `make gen-errors`.
# DO NOT MODIFY BY HAND.
#
# <ERRORS-AUTOGEN>
from .errors import (
    InternalServerError,
    UnsupportedFeatureError,
    ProtocolError,
    BinaryProtocolError,
    UnsupportedProtocolVersionError,
    TypeSpecNotFoundError,
    UnexpectedMessageError,
    InputDataError,
    ResultCardinalityMismatchError,
    CapabilityError,
    UnsupportedCapabilityError,
    DisabledCapabilityError,
    QueryError,
    InvalidSyntaxError,
    EdgeQLSyntaxError,
    SchemaSyntaxError,
    GraphQLSyntaxError,
    InvalidTypeError,
    InvalidTargetError,
    InvalidLinkTargetError,
    InvalidPropertyTargetError,
    InvalidReferenceError,
    UnknownModuleError,
    UnknownLinkError,
    UnknownPropertyError,
    UnknownUserError,
    UnknownDatabaseError,
    UnknownParameterError,
    SchemaError,
    SchemaDefinitionError,
    InvalidDefinitionError,
    InvalidModuleDefinitionError,
    InvalidLinkDefinitionError,
    InvalidPropertyDefinitionError,
    InvalidUserDefinitionError,
    InvalidDatabaseDefinitionError,
    InvalidOperatorDefinitionError,
    InvalidAliasDefinitionError,
    InvalidFunctionDefinitionError,
    InvalidConstraintDefinitionError,
    InvalidCastDefinitionError,
    DuplicateDefinitionError,
    DuplicateModuleDefinitionError,
    DuplicateLinkDefinitionError,
    DuplicatePropertyDefinitionError,
    DuplicateUserDefinitionError,
    DuplicateDatabaseDefinitionError,
    DuplicateOperatorDefinitionError,
    DuplicateViewDefinitionError,
    DuplicateFunctionDefinitionError,
    DuplicateConstraintDefinitionError,
    DuplicateCastDefinitionError,
    QueryTimeoutError,
    ExecutionError,
    InvalidValueError,
    DivisionByZeroError,
    NumericOutOfRangeError,
    IntegrityError,
    ConstraintViolationError,
    CardinalityViolationError,
    MissingRequiredError,
    TransactionError,
    TransactionConflictError,
    TransactionSerializationError,
    TransactionDeadlockError,
    ConfigurationError,
    AccessError,
    AuthenticationError,
    LogMessage,
    WarningMessage,
    ClientError,
    ClientConnectionError,
    ClientConnectionFailedError,
    ClientConnectionFailedTemporarilyError,
    ClientConnectionTimeoutError,
    ClientConnectionClosedError,
    InterfaceError,
    QueryArgumentError,
    MissingArgumentError,
    UnknownArgumentError,
    InvalidArgumentError,
    NoDataError,
)
# </ERRORS-AUTOGEN>



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2008-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import sys

if sys.version_info < (3, 6):
    # edgedb.NamedTuple() relies on ordered **kwargs.
    raise RuntimeError('edgedb requires Python 3.6 or greater')

import os
import os.path
import pathlib
import re
import subprocess

# We use vanilla build_ext, to avoid importing Cython via
# the setuptools version.
from distutils import extension as distutils_extension
from distutils.command import build_ext as distutils_build_ext

import setuptools
from setuptools.command import build_py as setuptools_build_py
from setuptools.command import sdist as setuptools_sdist


CYTHON_DEPENDENCY = 'Cython(>=0.29.24,<0.30.0)'

# Minimal dependencies required to test edgedb.
TEST_DEPENDENCIES = [
    # pycodestyle is a dependency of flake8, but it must be frozen because
    # their combination breaks too often
    # (example breakage: https://gitlab.com/pycqa/flake8/issues/427)
    'pycodestyle~=2.6.0',
    'pyflakes~=2.2.0',
    'flake8-bugbear~=21.4.3',
    'flake8~=3.8.1',
    'uvloop>=0.15.1; platform_system != "Windows" and python_version >= "3.7"',
]

# Dependencies required to build documentation.
DOC_DEPENDENCIES = [
    'sphinx~=2.3.1',
    'sphinxcontrib-asyncio~=0.2.0',
    'sphinx_rtd_theme~=0.4.3',
]

EXTRA_DEPENDENCIES = {
    'docs': DOC_DEPENDENCIES,
    'test': TEST_DEPENDENCIES,
    # Dependencies required to develop edgedb.
    'dev': [
        CYTHON_DEPENDENCY,
        'pytest>=3.6.0',
    ] + DOC_DEPENDENCIES + TEST_DEPENDENCIES
}


CFLAGS = ['-O2']
LDFLAGS = []
SYSTEM = sys.platform

if SYSTEM != 'win32':
    CFLAGS.extend(['-std=gnu99', '-fsigned-char', '-Wall',
                   '-Wsign-compare', '-Wconversion'])

if SYSTEM == 'darwin':
    # Lots of warnings from the standard library on macOS 10.14
    CFLAGS.extend(['-Wno-nullability-completeness'])

_ROOT = pathlib.Path(__file__).parent


with open(str(_ROOT / 'README.rst')) as f:
    readme = f.read()


with open(str(_ROOT / 'edgedb' / '_version.py')) as f:
    for line in f:
        if line.startswith('__version__ ='):
            _, _, version = line.partition('=')
            VERSION = version.strip(" \n'\"")
            break
    else:
        raise RuntimeError(
            'unable to read the version from edgedb/_version.py')


if (_ROOT / '.git').is_dir() and 'dev' in VERSION:
    # This is a git checkout, use git to
    # generate a precise version.
    def git_commitish():
        env = {}
        v = os.environ.get('PATH')
        if v is not None:
            env['PATH'] = v

        git = subprocess.run(['git', 'rev-parse', 'HEAD'], env=env,
                             cwd=str(_ROOT), stdout=subprocess.PIPE)
        if git.returncode == 0:
            commitish = git.stdout.strip().decode('ascii')
        else:
            commitish = 'unknown'

        return commitish

    VERSION += '+' + git_commitish()[:7]


class VersionMixin:

    def _fix_version(self, filename):
        # Replace edgedb.__version__ with the actual version
        # of the distribution (possibly inferred from git).

        with open(str(filename)) as f:
            content = f.read()

        version_re = r"(.*__version__\s*=\s*)'[^']+'(.*)"
        repl = r"\1'{}'\2".format(self.distribution.metadata.version)
        content = re.sub(version_re, repl, content)

        with open(str(filename), 'w') as f:
            f.write(content)


class sdist(setuptools_sdist.sdist, VersionMixin):

    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)
        self._fix_version(pathlib.Path(base_dir) / 'edgedb' / '_version.py')


class build_py(setuptools_build_py.build_py, VersionMixin):

    def build_module(self, module, module_file, package):
        outfile, copied = super().build_module(module, module_file, package)

        if module == '_version' and package == 'edgedb':
            self._fix_version(outfile)

        return outfile, copied


class build_ext(distutils_build_ext.build_ext):

    user_options = distutils_build_ext.build_ext.user_options + [
        ('cython-always', None,
            'run cythonize() even if .c files are present'),
        ('cython-annotate', None,
            'Produce a colorized HTML version of the Cython source.'),
        ('cython-directives=', None,
            'Cython compiler directives'),
    ]

    def initialize_options(self):
        # initialize_options() may be called multiple times on the
        # same command object, so make sure not to override previously
        # set options.
        if getattr(self, '_initialized', False):
            return

        super(build_ext, self).initialize_options()

        if os.environ.get('EDGEDB_DEBUG'):
            self.cython_always = True
            self.cython_annotate = True
            self.cython_directives = "linetrace=True"
            self.define = 'PG_DEBUG,CYTHON_TRACE,CYTHON_TRACE_NOGIL'
            self.debug = True
        else:
            self.cython_always = False
            self.cython_annotate = None
            self.cython_directives = None
            self.debug = False

    def finalize_options(self):
        # finalize_options() may be called multiple times on the
        # same command object, so make sure not to override previously
        # set options.
        if getattr(self, '_initialized', False):
            return

        need_cythonize = self.cython_always
        cfiles = {}

        for extension in self.distribution.ext_modules:
            for i, sfile in enumerate(extension.sources):
                if sfile.endswith('.pyx'):
                    prefix, ext = os.path.splitext(sfile)
                    cfile = prefix + '.c'

                    if os.path.exists(cfile) and not self.cython_always:
                        extension.sources[i] = cfile
                    else:
                        if os.path.exists(cfile):
                            cfiles[cfile] = os.path.getmtime(cfile)
                        else:
                            cfiles[cfile] = 0
                        need_cythonize = True

        if need_cythonize:
            import pkg_resources

            # Double check Cython presence in case setup_requires
            # didn't go into effect (most likely because someone
            # imported Cython before setup_requires injected the
            # correct egg into sys.path.
            try:
                import Cython
            except ImportError:
                raise RuntimeError(
                    'please install {} to compile edgedb from source'.format(
                        CYTHON_DEPENDENCY))

            cython_dep = pkg_resources.Requirement.parse(CYTHON_DEPENDENCY)
            if Cython.__version__ not in cython_dep:
                raise RuntimeError(
                    'edgedb requires {}, got Cython=={}'.format(
                        CYTHON_DEPENDENCY, Cython.__version__
                    ))

            from Cython.Build import cythonize

            directives = {
                'language_level': '3'
            }
            if self.cython_directives:
                for directive in self.cython_directives.split(','):
                    k, _, v = directive.partition('=')
                    if v.lower() == 'false':
                        v = False
                    if v.lower() == 'true':
                        v = True

                    directives[k] = v

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=directives,
                annotate=self.cython_annotate)

        super(build_ext, self).finalize_options()


INCLUDE_DIRS = [
    'edgedb/pgproto/',
    'edgedb/datatypes',
]


setup_requires = []

if (not (_ROOT / 'edgedb' / 'protocol' / 'protocol.c').exists() or
        '--cython-always' in sys.argv):
    # No Cython output, require Cython to build.
    setup_requires.append(CYTHON_DEPENDENCY)


with open(str(_ROOT / 'README.rst')) as f:
    readme = f.read()


setuptools.setup(
    name='edgedb',
    version=VERSION,
    description='EdgeDB Python driver',
    long_description=readme,
    platforms=['macOS', 'POSIX', 'Windows'],
    author='MagicStack Inc',
    author_email='hello@magic.io',
    url='https://github.com/edgedb/edgedb-python',
    license='Apache License, Version 2.0',
    packages=['edgedb'],
    provides=['edgedb'],
    zip_safe=False,
    include_package_data=True,
    package_data={'edgedb': ['py.typed']},
    ext_modules=[
        distutils_extension.Extension(
            "edgedb.pgproto.pgproto",
            ["edgedb/pgproto/pgproto.pyx"],
            extra_compile_args=CFLAGS,
            extra_link_args=LDFLAGS),

        distutils_extension.Extension(
            "edgedb.datatypes.datatypes",
            ["edgedb/datatypes/args.c",
             "edgedb/datatypes/record_desc.c",
             "edgedb/datatypes/tuple.c",
             "edgedb/datatypes/namedtuple.c",
             "edgedb/datatypes/object.c",
             "edgedb/datatypes/set.c",
             "edgedb/datatypes/hash.c",
             "edgedb/datatypes/array.c",
             "edgedb/datatypes/link.c",
             "edgedb/datatypes/linkset.c",
             "edgedb/datatypes/repr.c",
             "edgedb/datatypes/comp.c",
             "edgedb/datatypes/datatypes.pyx"],
            extra_compile_args=CFLAGS,
            extra_link_args=LDFLAGS),

        distutils_extension.Extension(
            "edgedb.protocol.protocol",
            ["edgedb/protocol/protocol.pyx"],
            extra_compile_args=CFLAGS,
            extra_link_args=LDFLAGS,
            include_dirs=INCLUDE_DIRS),

        distutils_extension.Extension(
            "edgedb.protocol.asyncio_proto",
            ["edgedb/protocol/asyncio_proto.pyx"],
            extra_compile_args=CFLAGS,
            extra_link_args=LDFLAGS,
            include_dirs=INCLUDE_DIRS),

        distutils_extension.Extension(
            "edgedb.protocol.blocking_proto",
            ["edgedb/protocol/blocking_proto.pyx"],
            extra_compile_args=CFLAGS,
            extra_link_args=LDFLAGS,
            include_dirs=INCLUDE_DIRS),
    ],
    cmdclass={'build_ext': build_ext},
    test_suite='tests.suite',
    install_requires=[
        'typing-extensions>=3.10.0',
        'certifi>=2021.5.30; platform_system == "Windows"',
    ],
    extras_require=EXTRA_DEPENDENCIES,
    setup_requires=setup_requires,
)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import pickle
import random
import unittest

import uuid
from uuid import UUID as std_UUID
from edgedb.protocol.protocol import UUID as c_UUID


special_uuids = frozenset({
    std_UUID('00000000-0000-0000-0000-000000000000'),
    std_UUID('00000000-0000-0000-0000-000000000001'),
    std_UUID('10000000-0000-0000-0000-000000000000'),
    std_UUID('10000000-0000-0000-0000-000000000001'),
    std_UUID('FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF'),
    std_UUID('0F0F0F0F-0F0F-0F0F-0F0F-0F0F0F0F0F0F'),
    std_UUID('F0F0F0F0-F0F0-F0F0-F0F0-F0F0F0F0F0F0'),
})

test_uuids = tuple(
    special_uuids |
    frozenset({uuid.uuid4() for _ in range(1000)})
)


class TestUuid(unittest.TestCase):

    def ensure_equal(self, uuid1, uuid2):
        self.assertEqual(uuid1.bytes_le, uuid2.bytes_le)
        self.assertEqual(uuid1.clock_seq, uuid2.clock_seq)
        self.assertEqual(uuid1.fields, uuid2.fields)
        self.assertEqual(uuid1.bytes, uuid2.bytes)
        self.assertEqual(uuid1.int, uuid2.int)
        self.assertEqual(uuid1.hex, uuid2.hex)
        self.assertEqual(uuid1.variant, uuid2.variant)
        self.assertEqual(uuid1.version, uuid2.version)
        self.assertEqual(uuid1.urn, uuid2.urn)
        self.assertEqual(uuid1, uuid2)
        self.assertEqual(int(uuid1), int(uuid2))
        self.assertEqual(str(uuid1), str(uuid2))
        self.assertEqual(repr(uuid1), repr(uuid2))
        self.assertEqual(hash(uuid1), hash(uuid2))
        self.assertGreaterEqual(uuid1, uuid2)
        self.assertGreaterEqual(uuid2, uuid1)
        self.assertLessEqual(uuid1, uuid2)
        self.assertLessEqual(uuid2, uuid1)

    def test_uuid_ctr_01(self):
        with self.assertRaisesRegex(ValueError, r'invalid UUID.*got 4'):
            c_UUID('test')

        with self.assertRaisesRegex(ValueError,
                                    r'invalid UUID.*decodes to less'):
            c_UUID('49e3b4e4-4761-11e9-9160-2f38d067497')

        for v in {'49e3b4e4476111e991602f38d067497aaaaa',
                  '49e3b4e4476111e991602f38d067497aaaa',
                  '49e3b4e4476111e991602f38d067497aaa',
                  '49e3b4e4476111e991602f38d067497aa',
                  '49e3b4e4476111e-991602f-38d067497aa'}:
            with self.assertRaisesRegex(ValueError,
                                        r'invalid UUID.*decodes to more'):
                print(c_UUID(v))

        with self.assertRaisesRegex(ValueError,
                                    r"invalid UUID.*unexpected.*'x'"):
            c_UUID('49e3b4e4-4761-11e9-9160-2f38dx67497a')

        with self.assertRaisesRegex(ValueError,
                                    r"invalid UUID.*unexpected"):
            c_UUID('49e3b4e4-4761-11160-2f😱3867497a')

        with self.assertRaisesRegex(ValueError,
                                    r"invalid UUID.*unexpected"):
            c_UUID('49e3b4e4-4761-11eE-\xAA60-2f38dx67497a')

    def test_uuid_ctr_02(self):
        for py_u in test_uuids:
            c_u = c_UUID(py_u.bytes)
            self.ensure_equal(py_u, c_u)

        for py_u in test_uuids:
            c_u = c_UUID(str(py_u))
            self.ensure_equal(py_u, c_u)

    def test_uuid_pickle(self):
        u = c_UUID('de197476-4763-11e9-91bf-7311c6dc588e')
        d = pickle.dumps(u)
        u2 = pickle.loads(d)
        self.assertEqual(u, u2)
        self.assertEqual(str(u), str(u2))
        self.assertEqual(u.bytes, u2.bytes)

    def test_uuid_instance(self):
        u = c_UUID('de197476-4763-11e9-91bf-7311c6dc588e')
        self.assertTrue(isinstance(u, uuid.UUID))
        self.assertTrue(issubclass(c_UUID, uuid.UUID))

    def test_uuid_compare(self):
        u = c_UUID('de197476-4763-11e9-91bf-7311c6dc588e')
        us = uuid.UUID(bytes=u.bytes)

        for us2 in test_uuids:
            u2 = c_UUID(us2.bytes)

            if us < us2:
                self.assertLess(u, u2)
                self.assertGreater(u2, u)
            elif us > us2:
                self.assertGreater(u, u2)
                self.assertLess(u2, u)

        u3 = c_UUID('de197476-4763-11e9-91bf-7311c6dc588e')
        self.assertTrue(u == u3)
        self.assertFalse(u != u3)
        self.assertTrue(u >= u3)
        self.assertTrue(u <= u3)

        a = c_UUID('10000000-0000-0000-0000-000000000001')
        b = c_UUID('10000000-0000-0000-0000-000000000000')
        self.assertGreater(a, b)
        self.assertLess(b, a)

    def test_uuid_comp(self):
        for _ in range(1000):
            ll = random.choice(test_uuids)
            rr = random.choice(test_uuids)

            # Testing comparison operators.
            # Explicitly use > and not assertGreater;
            # we want all comparisons to be explicit in this test.

            if ll > rr:
                self.assertTrue(c_UUID(ll.bytes) > rr)
                self.assertTrue(c_UUID(ll.bytes) > c_UUID(rr.bytes))
                self.assertTrue(ll > c_UUID(rr.bytes))

            if ll < rr:
                self.assertTrue(c_UUID(ll.bytes) < rr)
                self.assertTrue(c_UUID(ll.bytes) < c_UUID(rr.bytes))
                self.assertTrue(ll < c_UUID(rr.bytes))

            if ll >= rr:
                self.assertTrue(c_UUID(ll.bytes) >= rr)
                self.assertTrue(c_UUID(ll.bytes) >= c_UUID(rr.bytes))
                self.assertTrue(ll >= c_UUID(rr.bytes))

            if ll <= rr:
                self.assertTrue(c_UUID(ll.bytes) <= rr)
                self.assertTrue(c_UUID(ll.bytes) <= c_UUID(rr.bytes))
                self.assertTrue(ll <= c_UUID(rr.bytes))

            if ll != rr:
                self.assertTrue(c_UUID(ll.bytes) != rr)
                self.assertTrue(c_UUID(ll.bytes) != c_UUID(rr.bytes))
                self.assertTrue(ll != c_UUID(rr.bytes))

            if ll == rr:
                self.assertTrue(c_UUID(ll.bytes) == rr)
                self.assertTrue(c_UUID(ll.bytes) == c_UUID(rr.bytes))
                self.assertTrue(ll == c_UUID(rr.bytes))

            self.assertTrue(c_UUID(ll.bytes) >= ll)
            self.assertTrue(c_UUID(ll.bytes) <= ll)

            self.assertTrue(c_UUID(ll.bytes) == ll)
            self.assertTrue(c_UUID(ll.bytes) == ll)



import uuid
from uuid import UUID as std_UUID
from edgedb.protocol.protocol import UUID as c_UUID

import time


TEST_UUID = uuid.uuid4()
TEST_CUUID = c_UUID(TEST_UUID.bytes)

ubytes = uuid.uuid4().bytes
ustr = str(std_UUID(bytes=ubytes))

print(str(std_UUID(bytes=ubytes)))
print(str(c_UUID(ubytes)))

N = 1_000_000


assert issubclass(c_UUID, std_UUID)
assert isinstance(c_UUID(ubytes), std_UUID)
assert c_UUID(ubytes).bytes == std_UUID(bytes=ubytes).bytes, \
    f'{ubytes}: {c_UUID(ubytes).bytes}'
assert c_UUID(ubytes).hex == std_UUID(bytes=ubytes).hex
assert c_UUID(ubytes).int == std_UUID(bytes=ubytes).int
assert c_UUID(str(std_UUID(bytes=ubytes))).int == std_UUID(bytes=ubytes).int
assert str(c_UUID(ubytes)) == str(std_UUID(bytes=ubytes))
assert hash(c_UUID(ubytes)) == hash(std_UUID(bytes=ubytes))


print()

print(repr(std_UUID(bytes=ubytes)))
print(repr(c_UUID(ubytes)))


print()

st = time.monotonic()
for _ in range(N):
    std_UUID(bytes=ubytes)
std_total = time.monotonic() - st
print(f'std_UUID(bytes):\t  {std_total:.4f}')

st = time.monotonic()
for _ in range(N):
    c_UUID(ubytes)
c_total = time.monotonic() - st
print(f'c_UUID(bytes):\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')

st = time.monotonic()
for _ in range(N):
    object()
print(f'object():\t\t  {time.monotonic() - st:.4f}')


print()

st = time.monotonic()
for _ in range(N):
    std_UUID(ustr)
std_total = time.monotonic() - st
print(f'std_UUID(str):\t\t  {std_total:.4f}')

st = time.monotonic()
for _ in range(N):
    c_UUID(ustr)
c_total = time.monotonic() - st
print(f'c_UUID(str):\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')


print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    str(u)
std_total = time.monotonic() - st
print(f'str(std_UUID()):\t  {std_total:.4f}')

u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    str(u)
c_total = time.monotonic() - st
print(f'str(c_UUID()):\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')

u = object()
st = time.monotonic()
for _ in range(N):
    str(u)
print(f'str(object()):\t\t  {time.monotonic() - st:.4f}')


print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    u.bytes
std_total = time.monotonic() - st
print(f'std_UUID().bytes:\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    u.bytes
c_total = time.monotonic() - st
print(f'c_UUID().bytes:\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')

print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    u.int
std_total = time.monotonic() - st
print(f'std_UUID().int:\t\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    u.int
c_total = time.monotonic() - st
print(f'c_UUID().int:\t\t* {c_total:.4f}')

print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    u.hex
std_total = time.monotonic() - st
print(f'std_UUID().hex:\t\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    u.hex
c_total = time.monotonic() - st
print(f'c_UUID().hex:\t\t* {c_total:.4f}')

print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    hash(u)
std_total = time.monotonic() - st
print(f'hash(std_UUID()):\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    hash(u)
c_total = time.monotonic() - st
print(f'hash(c_UUID()):\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')

print()

dct = {_: _ for _ in range(1000)}

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    dct.get(u)
std_total = time.monotonic() - st
print(f'dct[std_UUID()]:\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    dct.get(u)
c_total = time.monotonic() - st
print(f'dct[c_UUID()]:\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')

print()

u = std_UUID(bytes=ubytes)
st = time.monotonic()
for _ in range(N):
    _ = u == TEST_UUID
std_total = time.monotonic() - st
print(f'std_UUID() ==:\t\t  {std_total:.4f}')


u = c_UUID(ubytes)
st = time.monotonic()
for _ in range(N):
    _ = u == TEST_CUUID
c_total = time.monotonic() - st
print(f'c_UUID() ==:\t\t* {c_total:.4f} ({std_total / c_total:.2f}x)')



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Helpers for SCRAM authentication."""

import base64
import hashlib
import hmac
import os
import typing

from .saslprep import saslprep


RAW_NONCE_LENGTH = 18

# Per recommendations in RFC 7677.
DEFAULT_SALT_LENGTH = 16
DEFAULT_ITERATIONS = 4096


def generate_salt(length: int = DEFAULT_SALT_LENGTH) -> bytes:
    return os.urandom(length)


def generate_nonce(length: int = RAW_NONCE_LENGTH) -> str:
    return B64(os.urandom(length))


def build_verifier(password: str, *, salt: typing.Optional[bytes] = None,
                   iterations: int = DEFAULT_ITERATIONS) -> str:
    """Build the SCRAM verifier for the given password.

    Returns a string in the following format:

        "<MECHANISM>$<iterations>:<salt>$<StoredKey>:<ServerKey>"

    The salt and keys are base64-encoded values.
    """
    password = saslprep(password).encode('utf-8')

    if salt is None:
        salt = generate_salt()

    salted_password = get_salted_password(password, salt, iterations)
    client_key = get_client_key(salted_password)
    stored_key = H(client_key)
    server_key = get_server_key(salted_password)

    return (f'SCRAM-SHA-256${iterations}:{B64(salt)}$'
            f'{B64(stored_key)}:{B64(server_key)}')


class SCRAMVerifier(typing.NamedTuple):

    mechanism: str
    iterations: int
    salt: bytes
    stored_key: bytes
    server_key: bytes


def parse_verifier(verifier: str) -> SCRAMVerifier:

    parts = verifier.split('$')
    if len(parts) != 3:
        raise ValueError('invalid SCRAM verifier')

    mechanism = parts[0]
    if mechanism != 'SCRAM-SHA-256':
        raise ValueError('invalid SCRAM verifier')

    iterations, _, salt = parts[1].partition(':')
    stored_key, _, server_key = parts[2].partition(':')
    if not salt or not server_key:
        raise ValueError('invalid SCRAM verifier')

    try:
        iterations = int(iterations)
    except ValueError:
        raise ValueError('invalid SCRAM verifier') from None

    return SCRAMVerifier(
        mechanism=mechanism,
        iterations=iterations,
        salt=base64.b64decode(salt),
        stored_key=base64.b64decode(stored_key),
        server_key=base64.b64decode(server_key),
    )


def parse_client_first_message(resp: bytes):

    # Relevant bits of RFC 5802:
    #
    #    saslname        = 1*(value-safe-char / "=2C" / "=3D")
    #                      ;; Conforms to <value>.
    #
    #    authzid         = "a=" saslname
    #                      ;; Protocol specific.
    #
    #    cb-name         = 1*(ALPHA / DIGIT / "." / "-")
    #                       ;; See RFC 5056, Section 7.
    #                       ;; E.g., "tls-server-end-point" or
    #                       ;; "tls-unique".
    #
    #    gs2-cbind-flag  = ("p=" cb-name) / "n" / "y"
    #                      ;; "n" -> client doesn't support channel binding.
    #                      ;; "y" -> client does support channel binding
    #                      ;;        but thinks the server does not.
    #                      ;; "p" -> client requires channel binding.
    #                      ;; The selected channel binding follows "p=".
    #
    #    gs2-header      = gs2-cbind-flag "," [ authzid ] ","
    #                      ;; GS2 header for SCRAM
    #                      ;; (the actual GS2 header includes an optional
    #                      ;; flag to indicate that the GSS mechanism is not
    #                      ;; "standard", but since SCRAM is "standard", we
    #                      ;; don't include that flag).
    #
    #    username        = "n=" saslname
    #                      ;; Usernames are prepared using SASLprep.
    #
    #    reserved-mext  = "m=" 1*(value-char)
    #                      ;; Reserved for signaling mandatory extensions.
    #                      ;; The exact syntax will be defined in
    #                      ;; the future.
    #
    #    nonce           = "r=" c-nonce [s-nonce]
    #                      ;; Second part provided by server.
    #
    #    c-nonce         = printable
    #
    #    client-first-message-bare =
    #                      [reserved-mext ","]
    #                      username "," nonce ["," extensions]
    #
    #    client-first-message =
    #                      gs2-header client-first-message-bare

    attrs = resp.split(b',')

    cb_attr = attrs[0]
    if cb_attr == b'y':
        cb = True
    elif cb_attr == b'n':
        cb = False
    elif cb_attr[0:1] == b'p':
        _, _, cb = cb_attr.partition(b'=')
        if not cb:
            raise ValueError('malformed SCRAM message')
    else:
        raise ValueError('malformed SCRAM message')

    authzid_attr = attrs[1]
    if authzid_attr:
        if authzid_attr[0:1] != b'a':
            raise ValueError('malformed SCRAM message')
        _, _, authzid = authzid_attr.partition(b'=')
    else:
        authzid = None

    user_attr = attrs[2]
    if user_attr[0:1] == b'm':
        raise ValueError('unsupported SCRAM extensions in message')
    elif user_attr[0:1] != b'n':
        raise ValueError('malformed SCRAM message')

    _, _, user = user_attr.partition(b'=')

    nonce_attr = attrs[3]
    if nonce_attr[0:1] != b'r':
        raise ValueError('malformed SCRAM message')

    _, _, nonce_bin = nonce_attr.partition(b'=')
    nonce = nonce_bin.decode('ascii')
    if not nonce.isprintable():
        raise ValueError('invalid characters in client nonce')

    # ["," extensions] are ignored

    return len(cb_attr) + 2, cb, authzid, user, nonce


def parse_client_final_message(
        msg: bytes, client_nonce: str, server_nonce: str):

    # Relevant bits of RFC 5802:
    #
    #   gs2-header       = gs2-cbind-flag "," [ authzid ] ","
    #                       ;; GS2 header for SCRAM
    #                       ;; (the actual GS2 header includes an optional
    #                       ;; flag to indicate that the GSS mechanism is not
    #                       ;; "standard", but since SCRAM is "standard", we
    #                       ;; don't include that flag).
    #
    #   cbind-input     = gs2-header [ cbind-data ]
    #                       ;; cbind-data MUST be present for
    #                       ;; gs2-cbind-flag of "p" and MUST be absent
    #                       ;; for "y" or "n".
    #
    #   channel-binding = "c=" base64
    #                       ;; base64 encoding of cbind-input.
    #
    #   proof           = "p=" base64
    #
    #   client-final-message-without-proof =
    #                       channel-binding "," nonce [","
    #                       extensions]
    #
    #   client-final-message =
    #                       client-final-message-without-proof "," proof

    attrs = msg.split(b',')

    cb_attr = attrs[0]
    if cb_attr[0:1] != b'c':
        raise ValueError('malformed SCRAM message')

    _, _, cb_data = cb_attr.partition(b'=')

    nonce_attr = attrs[1]
    if nonce_attr[0:1] != b'r':
        raise ValueError('malformed SCRAM message')

    _, _, nonce_bin = nonce_attr.partition(b'=')
    nonce = nonce_bin.decode('ascii')

    expected_nonce = f'{client_nonce}{server_nonce}'

    if nonce != expected_nonce:
        raise ValueError(
            'invalid SCRAM client-final message: nonce does not match')

    proof = None

    for attr in attrs[2:]:
        if attr[0:1] == b'p':
            _, _, proof = attr.partition(b'=')
            proof_attr_len = len(attr)
            proof = base64.b64decode(proof)
        elif proof is not None:
            raise ValueError('malformed SCRAM message')

    if proof is None:
        raise ValueError('malformed SCRAM message')

    return cb_data, proof, proof_attr_len + 1


def build_client_first_message(client_nonce: str, username: str) -> str:

    bare = f'n={saslprep(username)},r={client_nonce}'
    return f'n,,{bare}', bare


def build_server_first_message(server_nonce: str, client_nonce: str,
                               salt: bytes, iterations: int) -> str:

    return (
        f'r={client_nonce}{server_nonce},'
        f's={B64(salt)},i={iterations}'
    )


def build_auth_message(
        client_first_bare: bytes,
        server_first: bytes, client_final: bytes) -> bytes:

    return b'%b,%b,%b' % (client_first_bare, server_first, client_final)


def build_client_final_message(
        password: str,
        salt: bytes,
        iterations: int,
        client_first_bare: bytes,
        server_first: bytes,
        server_nonce: str) -> str:

    client_final = f'c=biws,r={server_nonce}'

    AuthMessage = build_auth_message(
        client_first_bare, server_first, client_final.encode('utf-8'))

    SaltedPassword = get_salted_password(
        saslprep(password).encode('utf-8'),
        salt,
        iterations)

    ClientKey = get_client_key(SaltedPassword)
    StoredKey = H(ClientKey)
    ClientSignature = HMAC(StoredKey, AuthMessage)
    ClientProof = XOR(ClientKey, ClientSignature)

    ServerKey = get_server_key(SaltedPassword)
    ServerProof = HMAC(ServerKey, AuthMessage)

    return f'{client_final},p={B64(ClientProof)}', ServerProof


def build_server_final_message(
        client_first_bare: bytes, server_first: bytes,
        client_final: bytes, server_key: bytes) -> str:

    AuthMessage = build_auth_message(
        client_first_bare, server_first, client_final)
    ServerSignature = HMAC(server_key, AuthMessage)
    return f'v={B64(ServerSignature)}'


def parse_server_first_message(msg: bytes):

    attrs = msg.split(b',')

    nonce_attr = attrs[0]
    if nonce_attr[0:1] != b'r':
        raise ValueError('malformed SCRAM message')

    _, _, nonce_bin = nonce_attr.partition(b'=')
    nonce = nonce_bin.decode('ascii')
    if not nonce.isprintable():
        raise ValueError('malformed SCRAM message')

    salt_attr = attrs[1]
    if salt_attr[0:1] != b's':
        raise ValueError('malformed SCRAM message')

    _, _, salt_b64 = salt_attr.partition(b'=')
    salt = base64.b64decode(salt_b64)

    iter_attr = attrs[2]
    if iter_attr[0:1] != b'i':
        raise ValueError('malformed SCRAM message')

    _, _, iterations = iter_attr.partition(b'=')

    try:
        itercount = int(iterations)
    except ValueError:
        raise ValueError('malformed SCRAM message') from None

    return nonce, salt, itercount


def parse_server_final_message(msg: bytes):

    attrs = msg.split(b',')

    nonce_attr = attrs[0]
    if nonce_attr[0:1] != b'v':
        raise ValueError('malformed SCRAM message')

    _, _, signature_b64 = nonce_attr.partition(b'=')
    signature = base64.b64decode(signature_b64)

    return signature


def verify_password(password: bytes, verifier: str) -> bool:
    """Check the given password against a verifier.

    Returns True if the password is OK, False otherwise.
    """

    password = saslprep(password).encode('utf-8')
    v = parse_verifier(verifier)
    salted_password = get_salted_password(password, v.salt, v.iterations)
    computed_key = get_server_key(salted_password)
    return v.server_key == computed_key


def verify_client_proof(client_first: bytes, server_first: bytes,
                        client_final: bytes, StoredKey: bytes,
                        ClientProof: bytes) -> bool:
    AuthMessage = build_auth_message(client_first, server_first, client_final)
    ClientSignature = HMAC(StoredKey, AuthMessage)
    ClientKey = XOR(ClientProof, ClientSignature)
    return H(ClientKey) == StoredKey


def B64(val: bytes) -> str:
    """Return base64-encoded string representation of input binary data."""
    return base64.b64encode(val).decode()


def HMAC(key: bytes, msg: bytes) -> bytes:
    return hmac.new(key, msg, digestmod=hashlib.sha256).digest()


def XOR(a: bytes, b: bytes) -> bytes:
    if len(a) != len(b):
        raise ValueError('scram.XOR received operands of unequal length')
    xint = int.from_bytes(a, 'big') ^ int.from_bytes(b, 'big')
    return xint.to_bytes(len(a), 'big')


def H(s: bytes) -> bytes:
    return hashlib.sha256(s).digest()


def get_salted_password(password: bytes, salt: bytes,
                        iterations: int) -> bytes:
    # U1 := HMAC(str, salt + INT(1))
    H_i = U_i = HMAC(password, salt + b'\x00\x00\x00\x01')

    for _ in range(iterations - 1):
        U_i = HMAC(password, U_i)
        H_i = XOR(H_i, U_i)

    return H_i


def get_client_key(salted_password: bytes) -> bytes:
    return HMAC(salted_password, b'Client Key')


def get_server_key(salted_password: bytes) -> bytes:
    return HMAC(salted_password, b'Server Key')



# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from libc cimport math


cdef float4_encode(CodecContext settings, WriteBuffer buf, obj):
    cdef double dval = cpython.PyFloat_AsDouble(obj)
    cdef float fval = <float>dval
    if math.isinf(fval) and not math.isinf(dval):
        raise ValueError('value out of float32 range')

    buf.write_int32(4)
    buf.write_float(fval)


cdef float4_decode(CodecContext settings, FRBuffer *buf):
    cdef float f = hton.unpack_float(frb_read(buf, 4))
    return cpython.PyFloat_FromDouble(f)


cdef float8_encode(CodecContext settings, WriteBuffer buf, obj):
    cdef double dval = cpython.PyFloat_AsDouble(obj)
    buf.write_int32(8)
    buf.write_double(dval)


cdef float8_decode(CodecContext settings, FRBuffer *buf):
    cdef double f = hton.unpack_double(frb_read(buf, 8))
    return cpython.PyFloat_FromDouble(f)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


cdef class EnumDescriptor:

    def __init__(self, object tid, tuple labels):
        self.tid = tid
        index = {}
        for i, l in enumerate(labels):
            index[l] = i
        self.index = index
        self.labels = labels

    cdef get_index(self, EnumValue v):
        return self.index[v.label]


cdef class EnumValue:

    def __init__(self, EnumDescriptor desc, str label):
        self.desc = desc
        self.label = label

    cdef get_index(self):
        return self.desc.get_index(self)

    def __str__(self):
        return self.label

    def __repr__(self):
        return f'<EnumValue {self.label!r}>'

    property __tid__:
        def __get__(self):
            return self.desc.tid

    def __eq__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.label == (<EnumValue>other).label

    def __ne__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.label != (<EnumValue>other).label

    def __lt__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.get_index() < (<EnumValue>other).get_index()

    def __gt__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.get_index() > (<EnumValue>other).get_index()

    def __le__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.get_index() <= (<EnumValue>other).get_index()

    def __ge__(self, other):
        if not isinstance(other, EnumValue):
            return NotImplemented
        if self.desc.tid != (<EnumValue>other).desc.tid:
            return NotImplemented
        return self.get_index() >= (<EnumValue>other).get_index()

    def __hash__(self):
        return hash((self.desc.tid, self.label))
